function [ypr_Bayes, mse_Bayes] = LBK_1(xpr, X1, Y, mixture_model)
% Computes the estimates at the points <xpr> using observations Y taken at
% design X1.
% The structure mixture model defines the stochastic model
% Considers a constant model for the mean term 
%
% collect local variables
n     = size(X1,1);  % design size
npr   = size(xpr,2); % size of prediction grid

THETA = mixture_model.THETA;
lT    = size(THETA,2);
p     = mixture_model.p;
np    = mixture_model.np;
cov_type_main       = mixture_model.cov_type_main;
cov_type_nonstat    = mixture_model.cov_type_nonstat;
Prior               = mixture_model.Prior;

nu0         = 1; % parameters for inverse chi-2 prior distribution of sig2
s20         = 1; 
F0          = ones(n,1); 
fx          = 1;
n0          = n;

ypr_Bayes = zeros(npr,1); % returned estimates
mse_Bayes = zeros(npr,1); % mean square error of returned estimates

% these vectors store the results for each mixture component
Yx          = zeros(1,lT);
Mse         = zeros(1,lT);
Posterior   = Prior;

for i=1:npr
    
    for k=1:lT      % goes over the mixture components
        theta_main      =   THETA(1,k);
        theta_nonstat   =   THETA(2,k);
        % Prediction
        rx0     = calcrx(X1',theta_main,p,xpr(i),cov_type_main); 
        R0      = calcR(X1',theta_main,p,cov_type_main);
        rx1     = calcrx(X1',theta_nonstat,p,xpr(i),cov_type_nonstat);
        Dm1     = diag(sqrt(rx1));
        F       = Dm1*F0;
        beta    = ((F'/R0)*F)\((F'/R0)*Dm1*Y);
        Yx(k)   = fx*beta+rx0'*(R0\(Dm1*(Y-F0*beta)));  
        % Likelihood
        n_equiv = n0;
        sig2    = (1/(n_equiv-np))*(Y-F0*beta)'*Dm1*(R0\(Dm1*(Y-F0*beta)));
        % Expected Likelihood with a uniform prior for beta and inv-chi-2 for sig2
        nu0p    = nu0+n_equiv-np; 
        s20p    = (nu0*s20+(n_equiv-np)*sig2)/(nu0+n_equiv-np);
        ELn     = (det(R0))^(-1/2)*prod(sqrt(rx1))*(det((F'/R0)*F))^(-1/2)*(2*pi)^(-(n_equiv-np)/2)*...
                gamma(nu0p/2)/gamma(nu0/2)*((s20*nu0/2)^(nu0/2))/((s20p*nu0p/2)^(nu0p/2));
        Posterior(k) = Prior(k)*ELn;
       
        n_equiv = sum(rx1);   % accounts for the small number of effective 
                              % observations due to non-stationarity :
                              % only used to compute mse
        sig2    = (n0-np)*sig2/(n_equiv-np);
        nu0p    = nu0+n_equiv-np; 
        s20p    = (nu0*s20+(n_equiv-np)*sig2)/(nu0+n_equiv-np);
            
        Mse(k) = s20p*(1-rx0'*(R0\rx0) + (1-rx0'*(R0\F))^2/((F'/R0)*F) );
    end

    Posterior       = Posterior/(sum(Posterior));
    ypr_Bayes(i)    = Yx*Posterior';
    mse_Bayes(i)    = max(0, Mse*Posterior'+((Yx-Yx*Posterior').^2)*Posterior');
end
end


%%================
function rx=calcrx(X,theta,p,x,cov_type)
% function rx = calcrx(X,theta,p,x,cov_type) 
% computes the correlations between predicted and design points
%
% x     : point where field is predicted : y(x)
% rx    : vector of correlations between predicted and design points
% theta & p : correlation function parameters 
% X     : design matrix (each column is a design point x(i), i = 1,..., n)
% cov_type : selects the covariance kernel 
% cov_type==1 : exponential family + spherical
% cov_type==2 : exponential family, product type (p in [1,2])
% cov_type==3 : spherical bounded support
% cov_type==4 : Stein 1
% cov_type==5 : Matern 3/2 + spherical (p=1)
% cov_type==6 : Matern 5/2 + spherical (p=1)
% cov_type==7 : Non stationary [Rasmussen, 2006, p 93]

[dimx,n]=size(X);

if cov_type==1 % exponential family + spherical    
    S  = sqrt(ones(1,dimx)*(abs(X-x*ones(1,n))).^2);
    rx = exp(-theta*S.^p)';
    
elseif cov_type==2 % exponential family, product type (p in [1,2])
    rx = exp(-theta'*(abs(X-x*ones(1,n))).^p);
    
elseif cov_type==3 %spherical bounded support
    dd = sqrt(sum((X-x*ones(1,n)).^2))'; 
    rx = spherical_cov( dd );
     
elseif cov_type==4  % Stein 1
     if dimx>1
        pr = prod( ones(dimx,n)+(theta*ones(1,n)).*abs(X-x*ones(1,n)) );
     else
        pr = ones(dimx,n)+(theta*ones(1,n)).*abs(X-x*ones(1,n));
     end
     rx = exp(-theta'*(abs(X-x*ones(1,n))).^p);
     rx = (rx.*pr)';
     
elseif cov_type==5  % Matern 3/2 + spherical (p=1)
     S  = sqrt(ones(1,dimx)*(abs(X-x*ones(1,n))).^2);
     rx = ((1+theta*S).*exp(-theta*S))';
     
elseif cov_type==6  % Matern 5/2 + spherical (p=1)
     S  = sqrt(ones(1,dimx)*(abs(X-x*ones(1,n))).^2);
     rx = ((1+theta*S+theta^2*S.^2/3).*exp(-theta*S))';
     
elseif cov_type==7 % Non stationary [Rasmussen, 2006, p 93]
     x0 = 3*ones(dimx,1);
     LX = 1-0.75*exp(-8*(abs(X-x0*ones(1,n)).^2)); % dimx times n
     Lx = 1-0.75*exp(-8*(abs(x-x0).^2));           % dimx times 1
     if dimx>1 
         cc = prod(sqrt(2*LX.*(Lx*ones(1,n))./(LX.^2+(Lx*ones(1,n)).^2)));
     else
         cc = sqrt(2*LX.*(Lx*ones(1,n))./(LX.^2+(Lx*ones(1,n)).^2));
     end
    S  = ones(1,dimx)*((abs(X-x*ones(1,n))).^2./(LX.^2+(Lx*ones(1,n)).^2));
    rx = (exp(-S).*cc)';
end

end
%%================
function R=calcR(X,theta,p,cov_type)
% function R=calcR(X,theta,p,cov_type) 
% R : matrice de corrlation des z(x(i)), i = 1,..., n
% theta & p : paramtres des corrlations R et rx
% global cov_type: % specifies the type of covariance structure
% X : matrice dont les colonnes sont les points x(i), i = 1,..., n

[dimx,n]=size(X);
R=eye(n)/2;
for i=1:n-1
    for j=i+1:n
        if cov_type==1 % exponential family + spherical
            S=sqrt(sum(abs(X(:,i)-X(:,j)).^2));
            R(i,j)=exp(-theta*S^p);
        elseif cov_type==2 % exponential family, product type (p in [1,2])
            R(i,j)=exp(-theta'*(abs(X(:,i)-X(:,j)).^p))
        elseif cov_type==3 %spherical bounded support
            dd=sqrt(sum((X(:,i)-X(:,j)).^2));
            R(i,j)=spherical_cov( dd );
        elseif cov_type==4  % Stein 1
            pr=prod( ones(dimx,1) + theta.*abs(X(:,i)-X(:,j)));
            R(i,j)=exp(-theta'*(abs(X(:,i)-X(:,j)).^p))*pr;
        elseif cov_type==5  % Matern 3/2 + spherical (p=1)
            % = 1/2 for theta*r=1.67835
            % = 1/5 for theta*r=2.994308347
            S=sqrt(sum(abs(X(:,i)-X(:,j)).^2));
            R(i,j)=(1+theta*S)*exp(-theta*S);
        elseif cov_type==6  % Matern 5/2 + spherical (p=1)
            % = 1/2 for theta*r=2.33026
            % = 1/5 for theta*r=3.914105418
            S=sqrt(sum(abs(X(:,i)-X(:,j)).^2));
            R(i,j)=(1+theta*S+theta^2*S.^2/3)*exp(-theta*S);
        elseif cov_type==7 % Non stationary [Rasmussen, 2006, p 93]
            x0=3*ones(dimx,1);
            Lxi=1-0.75*exp(-8*(abs(X(:,i)-x0).^2)); % dimx times 1
            Lxj=1-0.75*exp(-8*(abs(X(:,j)-x0).^2));
            cc=prod(sqrt(2*Lxi.*Lxj./(Lxi.^2+Lxj.^2)));
            S=sum(abs(X(:,i)-X(:,j)).^2./(Lxi.^2+Lxj.^2));
            R(i,j)=exp(-S)*cc;
        end
    end
end
R=R+triu(R)';
end
%%================
function [ LL ] = log_like( theta )
% function [ LL ] = log_like( theta )
% computes the  (restricted) neg-loglikelihood at theta

global cov_type_g p_g X0_g Z0_g F_g

[n0,np] = size(F_g);

R       = calcR(X0_g,theta,p_g,cov_type_g);
beta    = ((F_g'/R)*F_g)\((F_g'/R)*Z0_g);               % mean coefficients
sig2    = ((Z0_g-F_g*beta)'/R)*(Z0_g-F_g*beta)/(n0-np); % variance estimate
LL      = (n0-np)/2*log(sig2)+1/2*log(det(R));

end