% Local Bayesian Krigin implementation in dimension 1
% L.Pronzato (I3S/CNRS, France), 2015
% Code corresponds to article "Local Bayesian Kriging"
% Technometrics, 2016
%
% This script ilustrates the use of Local Bayesian Kriging in dimension
% one.
% The estimator is run on a random choice of the values of a simulated
% Gaussian Process or on a deterministic signal (see flag TRUE_SIMUL
% below).
% A final plot shows the true (simulated) process, the estimated signal
% and its 95% confidence bounds.
function Run_LBK_1()
clear all
hold off

% intialize random generators
% rand('seed',25)
% randn('seed',7)


DBG = 1;    % set to 0 stops verbose reporting and ploting

% the constants below are used for ploting the confidence intervals
CV_kriging = 3.182;     % critical value t 3 degrees of freedom (95%)
CV_Bayes   = 2.571;     % critical value t 5 degrees of freedom (95%)
	
TRUE_SIMUL    = 1  % if equal to zero a deterministic function is used instead of a simulated Gaussian Process.
USE_UNIFORM   = 1; % if equal to one the prior based on the distance between components is used
RANDOM_DESIGN = 1; % if equal to one a random design in the unit interval is used
%______________________________________________________
% 1) Simulation d'une trajectoire de processus
%______________________________________________________
% the following variable selects the covariance kernel:
% cov_type==1 : exponential family + spherical
% cov_type==2 : exponential family, product type (p in [1,2])
% cov_type==3 : spherical bounded support
% cov_type==4 : Stein 1
% cov_type==5 : Matern 3/2 + spherical (p=1)
% cov_type==6 : Matern 5/2 + spherical (p=1)
% cov_type==7 : Non stationary [Rasmussen, 2006, p 93]

%% simulation grid
N       = 200;	
lower   = -1; upper = 1;  % complete simulation grid
step    = (upper - lower)/(N-1);
X       = lower:step:upper;
length(X)

% simulated process
cov_type_simul = 5;
theta_true     = 2.; 
p              = 2;
s2=1.; sigma=sqrt(s2);

if TRUE_SIMUL==0
    % specify Z(X)
    Z = X.*(sin(10*X+1)+0.1*sin(15*X)); % Deceptive function of [Benassi, Bect, Vazquez, 2011]
else    
    XX      = X(1);
    Z       = sigma*randn(1,1);
    CN      = calcR(XX, theta_true, p, cov_type_simul);
    CCN     = chol(CN); 
    ICCN    = inv(CCN);
    ZpI     = Z'*ICCN;

    I_perm=randperm(N-1); % randomly incorporate the points in the sequential computation

    for i=1: N-1
        j   = I_perm(i)+1; % 1st point has been used in initialization
        x   = X(j);
        rx  = calcrx(XX,theta_true,p,x',cov_type_simul);
        r1p  = rx'*ICCN; r1=r1p';
        v1   = 1-r1'*r1; 		% similar to (1-rx'*inv(CN)*rx)
        a1   = sqrt(v1); 
        a1   = max(a1,0); 
        % mean m
        m    = ZpI*r1;
        m    = real(m);	
        % z(x)
        z    = m+sigma*a1*randn(1,1);

        if isinf(z)==0
            % update ICCN, inverse of cholevsky of CN=[CN rx;rx' 1];	
            n       = length(Z);
            ICCN    = [ICCN -ICCN*r1/a1; zeros(1,n) 1/a1];
            % update X Z
            XX       = [XX, x]; 
            Z       = [Z;z];
            % update ZpI (used in m)
            ZpI     = [ZpI (z-m)/a1];
        end
    end
    % re-order the generated observations
    [X, xs] = sort(XX); 
    Z       = Z(xs)';
end    
%
if (DBG)
    figure(1)
    plot(X,Z,'r-','LineWidth',2,'MarkerSize',12)
    xlabel('x','fontsize',14), ylabel('f(x)','fontsize',14)
    if TRUE_SIMUL == 0
        title('deterministic function')
    else
        title('simulated process')
    end
    pause(2) 
end

%% define the design 

if RANDOM_DESIGN == 1
    n = 10; % design size
    design = [];
    while (length(design) < n)
        design = randi([1, length(X)], 1, n); % to have the design points spaced at least of 4 grid points
        design = unique(design);
    end
else
    design = [1    10    20    50    80   100   200]; 
    n = length(design);
end
X_design = X(design)';
Z_design = Z(design);          % observations are a column vector
%% define the mixture 

% define the mixture model
mixture_model.THETA                 = [1 5 10 20 ;
                                       5*ones(1,4)];
mixture_model.p                     = p;
mixture_model.np                    = 1;
mixture_model.cov_type_main         = 5; % main kernel (can be different from the one used to simulate the sample path)
mixture_model.cov_type_nonstat      = 5; % kernel for non-stationarity 

lT = size(mixture_model.THETA, 2);
    
% prior weights
if (USE_UNIFORM == 1)
    mixture_model.Prior  = ones(1,lT)/lT;    % Prior weigths = uniform
else
    %----------------------------------------------------------------------
    % Calculation of prior according to [P.H. Garthwaite and E. Mubwandarikwa, Selection of weights for weighted model averaging,
    % Aust. N.Z.J. Stat., 52(4):363-382, 2010] based on a correlation amtrix
    % built from Hellinger distances between normal distributions
    KK  = zeros(lT,lT);
    for k1=1:lT-1
        Rk1 = calcR(X_design',mixture_model.THETA(1,k1),...
            mixture_model.p, mixture_model.cov_type_main);
        for k2=k1+1:lT
            Rk2       = calcR(X_design',mixture_model.THETA(1,k2),...
                mixture_model.p, mixture_model.cov_type_main);
            KK(k1,k2) = Hellinger_normals_0( Rk1, Rk2 );
            KK(k2,k1) = KK(k1,k2);
        end
    end
    mixture_model.Prior = real(cos_square_weighting( 1-KK.^2,1e-6 ))';
    if (DBG) 
       display(sprintf('computed prior'));
       display(sprintf(' %g', mixture_model.Prior)); 
    end
end

%% call the estimator
xpr  = X;               % grid for prediction

%

[ypr_Bayes, mse_Bayes] = LBK_1(xpr, X_design, Z_design', mixture_model);

if (DBG)
    figure(2); clf; 
    plot(X_design,Z_design,'rp',X,Z,'r-',...
        xpr,ypr_Bayes,'g-',... 
        xpr,ypr_Bayes+CV_Bayes*sqrt(mse_Bayes),'k--',xpr,ypr_Bayes-CV_Bayes*sqrt(mse_Bayes),'k--',...
        'LineWidth',2,'MarkerSize',12)
    legend('design','true','LBK estimate', 'confidence interval')
    xlabel('x','fontsize',14), ylabel('f(x)','fontsize',14)
    pause(2)
end
end
%% functions used in this script
%%=========================================================================
function w = DefineGridWeights(grid, a, tau1, tau2)
%%=========================================================================
% defines the weight for simulation of the non-stationary signal
fx1 = exp(a*(grid(1,:)-tau1))./(exp(a*(grid(1,:)-tau1))+exp(-a*(grid(1,:)-tau1)));
fx2 = exp(a*(grid(2,:)-tau2))./(exp(a*(grid(2,:)-tau2))+exp(-a*(grid(2,:)-tau2)));
w   = fx1.*fx2;

end

%%=========================================================================
function rx=calcrx(X,theta,p,x,cov_type)
%%=========================================================================
% function rx = calcrx(X,theta,p,x,cov_type) 
% computes the correlations between predicted and design points
%
% x     : point where field is predicted : y(x)
% rx    : vector of correlations between predicted and design points
% theta & p : correlation function parameters 
% X     : design matrix (each column is a design point x(i), i = 1,..., n)
% cov_type : selects the covariance kernel 
% cov_type==1 : exponential family + spherical
% cov_type==2 : exponential family, product type (p in [1,2])
% cov_type==3 : spherical bounded support
% cov_type==4 : Stein 1
% cov_type==5 : Matern 3/2 + spherical (p=1)
% cov_type==6 : Matern 5/2 + spherical (p=1)
% cov_type==7 : Non stationary [Rasmussen, 2006, p 93]

[dimx,n]=size(X);

if cov_type==1 % exponential family + spherical    
    S  = sqrt(ones(1,dimx)*(abs(X-x*ones(1,n))).^2);
    rx = exp(-theta*S.^p)';
    
elseif cov_type==2 % exponential family, product type (p in [1,2])
    rx = exp(-theta'*(abs(X-x*ones(1,n))).^p);
    
elseif cov_type==3 %spherical bounded support
    dd = sqrt(sum((X-x*ones(1,n)).^2))'; 
    rx = spherical_cov( dd );
     
elseif cov_type==4  % Stein 1
     if dimx>1
        pr = prod( ones(dimx,n)+(theta*ones(1,n)).*abs(X-x*ones(1,n)) );
     else
        pr = ones(dimx,n)+(theta*ones(1,n)).*abs(X-x*ones(1,n));
     end
     rx = exp(-theta'*(abs(X-x*ones(1,n))).^p);
     rx = (rx.*pr)';
     
elseif cov_type==5  % Matern 3/2 + spherical (p=1)
     S  = sqrt(ones(1,dimx)*(abs(X-x*ones(1,n))).^2);
     rx = ((1+theta*S).*exp(-theta*S))';
     
elseif cov_type==6  % Matern 5/2 + spherical (p=1)
     S  = sqrt(ones(1,dimx)*(abs(X-x*ones(1,n))).^2);
     rx = ((1+theta*S+theta^2*S.^2/3).*exp(-theta*S))';
     
elseif cov_type==7 % Non stationary [Rasmussen, 2006, p 93]
     x0 = 3*ones(dimx,1);
     LX = 1-0.75*exp(-8*(abs(X-x0*ones(1,n)).^2)); % dimx times n
     Lx = 1-0.75*exp(-8*(abs(x-x0).^2));           % dimx times 1
     if dimx>1 
         cc = prod(sqrt(2*LX.*(Lx*ones(1,n))./(LX.^2+(Lx*ones(1,n)).^2)));
     else
         cc = sqrt(2*LX.*(Lx*ones(1,n))./(LX.^2+(Lx*ones(1,n)).^2));
     end
    S  = ones(1,dimx)*((abs(X-x*ones(1,n))).^2./(LX.^2+(Lx*ones(1,n)).^2));
    rx = (exp(-S).*cc)';
end

end

%%=========================================================================
function R=calcR(X,theta,p,cov_type)
%%=========================================================================
% function R=calcR(X,theta,p,cov_type) 
% R : correlation matrix of the observations z(x(i)), i = 1,..., n
% theta & p : parameters of the correlation matrix R and vector rx
% cov_type_g % specifies the type of covariance structure
% X : matrix whose columns are the points x(i), i = 1,..., n

[dimx,n]=size(X);
R=eye(n)/2;
for i=1:n-1
    for j=i+1:n
        if cov_type==1 % exponential family + spherical
            S=sqrt(sum(abs(X(:,i)-X(:,j)).^2));
            R(i,j)=exp(-theta*S^p);
        elseif cov_type==2 % exponential family, product type (p in [1,2])
            R(i,j)=exp(-theta'*(abs(X(:,i)-X(:,j)).^p))
        elseif cov_type==3 %spherical bounded support
            dd=sqrt(sum((X(:,i)-X(:,j)).^2));
            R(i,j)=spherical_cov( dd );
        elseif cov_type==4  % Stein 1
            pr=prod( ones(dimx,1) + theta.*abs(X(:,i)-X(:,j)));
            R(i,j)=exp(-theta'*(abs(X(:,i)-X(:,j)).^p))*pr;
        elseif cov_type==5  % Matern 3/2 + spherical (p=1)
            S=sqrt(sum(abs(X(:,i)-X(:,j)).^2));
            R(i,j)=(1+theta*S)*exp(-theta*S);
        elseif cov_type==6  % Matern 5/2 + spherical (p=1)
            S=sqrt(sum(abs(X(:,i)-X(:,j)).^2));
            R(i,j)=(1+theta*S+theta^2*S.^2/3)*exp(-theta*S);
        elseif cov_type==7 % Non stationary [Rasmussen, 2006, p 93]
            x0=3*ones(dimx,1);
            Lxi=1-0.75*exp(-8*(abs(X(:,i)-x0).^2)); 
            Lxj=1-0.75*exp(-8*(abs(X(:,j)-x0).^2));
            cc=prod(sqrt(2*Lxi.*Lxj./(Lxi.^2+Lxj.^2)));
            S=sum(abs(X(:,i)-X(:,j)).^2./(Lxi.^2+Lxj.^2));
            R(i,j)=exp(-S)*cc;
        end
    end
end
R=R+triu(R)';

end
%%=========================================================================
% FUNCTIONS BELOW ARE USED FOR PRIOR COMPUTATION
%%=========================================================================
function [ H ] = Hellinger_normals_0( C1, C2 )
%%=========================================================================
% function [ H ] = Hellinger_normals_0( C1,C2 )
% C1 and C2 two covariance matrices n*n
% returns the Hellinger distance H(pi1,pi2) between the two normal distributions
% with same mean and covariances C1 and C2
BC=(det(C1*C2))^(1/4)/(det((C1+C2)/2))^(1/2); % Bhattacharyya coefficient
H=sqrt(1-BC);

end

%%=========================================================================
function [ w ] = cos_square_weighting( C,epsilon )
%%=========================================================================
% function [ w ] = cos_square_weighting( C,epsilon )
% C is a n*n correlation matrix (positive definite, 1's on the diagonal)
% returns n*1 vector of weigths according to the cos-square weighting
% scheme of [P.H. Garthwaite and E. Mubwandarikwa, Selection of weights for weighted model averaging, 
% Aust. N.Z.J. Stat., 52(4):363-382, 2010]
% epsilon (<1) sets the stopping rule in the iterative procedure : we stop when
% sum_{i=1,...n}|w_i^(k+1) - w_i^(k)|<epsilon, with w_i^(k) the ith weight
% at iteration k

[n,~]=size(C);
d=ones(n,1);
D=diag(d);
delta=1;
%k=0;
while delta>epsilon
    %k=k+1;
    [Q,Lambda]=eig(D*C*D);
    lambda=diag(Lambda);
    E=diag(d.^(-1))*Q*diag(lambda.^(1/2))*Q';
    dp=diag(E); 
    Dp=diag(dp);
    delta=sum(abs(d-dp));
    D=Dp; d=dp;
end
%k
w=D.^2; 
w=w/sum(w);
end


